@@ -8,16 +8,23 @@ from inspect import cleandoc
 from os.path import join
 from textwrap import TextWrapper
 
-from bundlewrap.exceptions import BundleError, ItemDependencyError, FaultUnavailable
-from bundlewrap.utils import cached_property
-from bundlewrap.utils.dicts import diff_keys, diff_value, hash_statedict, validate_statedict
+from bundlewrap.exceptions import (
+    BundleError,
+    FaultUnavailable,
+    ItemDependencyError,
+    ItemSkipped,
+)
+from bundlewrap.utils import cached_property, Fault
+from bundlewrap.utils.dicts import dict_to_text, diff_dict, hash_statedict, validate_statedict
 from bundlewrap.utils.text import force_text, mark_for_translation as _
-from bundlewrap.utils.text import blue, bold, italic, wrap_question
+from bundlewrap.utils.text import blue, bold, green, italic, red, wrap_question
 from bundlewrap.utils.ui import io
 from bundlewrap.operations import run_local
 
 
 BUILTIN_ITEM_ATTRIBUTES = {
+    'after': set(),
+    'before': set(),
     'cascade_skip': None,
     'comment': None,
     'needed_by': set(),
@@ -25,6 +32,7 @@ BUILTIN_ITEM_ATTRIBUTES = {
     'preceded_by': set(),
     'precedes': set(),
     'error_on_missing_fault': False,
+    'skip': False,
     'tags': set(),
     'triggered': False,
     'triggered_by': set(),
@@ -49,6 +57,18 @@ def format_comment(comment):
     return result
 
 
+def keys_to_fix(cdict, sdict):
+    if cdict is None:
+        return set()
+    if sdict is None:
+        return set(cdict.keys())
+    differing_keys = set()
+    for key, value in cdict.items():
+        if value != sdict[key]:
+            differing_keys.add(key)
+    return differing_keys
+
+
 class ItemStatus:
     """
     Holds information on a particular Item such as whether it needs
@@ -62,7 +82,7 @@ class ItemStatus:
         self.must_be_deleted = (self.sdict is not None and self.cdict is None)
         self.must_be_created = (self.cdict is not None and self.sdict is None)
         if not self.must_be_deleted and not self.must_be_created:
-            self.keys_to_fix = diff_keys(cdict, sdict)
+            self.keys_to_fix = keys_to_fix(cdict, sdict)
 
     def __repr__(self):
         return "<ItemStatus correct:{}>".format(self.correct)
@@ -95,6 +115,7 @@ class Item:
     BUNDLE_ATTRIBUTE_NAME = None
     ITEM_ATTRIBUTES = {}
     ITEM_TYPE_NAME = None
+    REJECT_UNKNOWN_ATTRIBUTES = True
     REQUIRED_ATTRIBUTES = []
     SKIP_REASON_CMDLINE = 1
     SKIP_REASON_DEP_FAILED = 2
@@ -105,6 +126,7 @@ class Item:
     SKIP_REASON_SOFTLOCK = 7
     SKIP_REASON_UNLESS = 8
     SKIP_REASON_DEP_SKIPPED = 9
+    SKIP_REASON_ATTR = 10
     SKIP_REASON_DESC = {
         SKIP_REASON_CMDLINE: _("cmdline"),
         SKIP_REASON_DEP_FAILED: _("dependency failed"),
@@ -115,6 +137,7 @@ class Item:
         SKIP_REASON_SOFTLOCK: _("soft locked"),
         SKIP_REASON_UNLESS: _("unless"),
         SKIP_REASON_DEP_SKIPPED: _("dependency skipped"),
+        SKIP_REASON_ATTR: _("attribute"),
     }
     STATUS_OK = 1
     STATUS_FIXED = 2
@@ -198,8 +221,20 @@ class Item:
             except FaultUnavailable:
                 self._faults_missing_for_attributes.add('when_creating/' + attribute_name)
 
+        if not self.REJECT_UNKNOWN_ATTRIBUTES:
+            for key, value in attributes.items():
+                if key in BUILTIN_ITEM_ATTRIBUTES:
+                    continue
+                if isinstance(value, Fault):
+                    try:
+                        value = value.value
+                    except FaultUnavailable:
+                        self._faults_missing_for_attributes.add(key)
+                        continue
+                self.attributes.setdefault(key, value)
+
         if self.cascade_skip is None:
-            self.cascade_skip = not (self.unless or self.triggered)
+            self.cascade_skip = not (self.skip or self.triggered or self.unless)
 
         if self.id in self.triggers:
             raise BundleError(_(
@@ -224,24 +259,16 @@ class Item:
     def __repr__(self):
         return "<Item {}>".format(self.id)
 
-    def _check_bundle_collisions(self, items):
-        for item in items:
-            if item == self:
-                continue
-            if item.id == self.id:
-                raise BundleError(_(
-                    "duplicate definition of {item} in bundles '{bundle1}' and '{bundle2}'"
-                ).format(
-                    item=item.id,
-                    bundle1=item.bundle.name,
-                    bundle2=self.bundle.name,
-                ))
-
     def _check_loopback_dependency(self):
         """
         Alerts the user if they have an item depend on itself.
         """
-        if self.id in self.needs or self.id in self.needed_by:
+        if (
+            self.id in self.after or
+            self.id in self.before or
+            self.id in self.needs or 
+            self.id in self.needed_by
+        ):
             raise ItemDependencyError(_(
                 "'{item}' in bundle '{bundle}' on node '{node}' cannot depend on itself"
             ).format(
@@ -369,34 +396,43 @@ class Item:
                 item=item_id,
                 bundle=bundle.name,
             ))
-        invalid_attributes = set(attributes.keys()).difference(
-            set(cls.ITEM_ATTRIBUTES.keys()).union(
-                set(BUILTIN_ITEM_ATTRIBUTES.keys())
-            ),
-        )
-        if invalid_attributes:
-            raise BundleError(_(
-                "invalid attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
-            ).format(
-                item=item_id,
-                bundle=bundle.name,
-                attrs=", ".join(invalid_attributes),
-            ))
 
-        invalid_attributes = set(attributes.get('when_creating', {}).keys()).difference(
-            set(cls.WHEN_CREATING_ATTRIBUTES.keys())
-        )
-        if invalid_attributes:
-            raise BundleError(_(
-                "invalid when_creating attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
-            ).format(
-                item=item_id,
-                bundle=bundle.name,
-                attrs=", ".join(invalid_attributes),
-            ))
+        if cls.REJECT_UNKNOWN_ATTRIBUTES:
+            invalid_attributes = set(attributes.keys()).difference(
+                set(cls.ITEM_ATTRIBUTES.keys()).union(
+                    set(BUILTIN_ITEM_ATTRIBUTES.keys())
+                ),
+            )
+            if invalid_attributes:
+                raise BundleError(_(
+                    "invalid attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
+                ).format(
+                    item=item_id,
+                    bundle=bundle.name,
+                    attrs=", ".join(invalid_attributes),
+                ))
+
+            invalid_attributes = set(attributes.get('when_creating', {}).keys()).difference(
+                set(cls.WHEN_CREATING_ATTRIBUTES.keys())
+            )
+            if invalid_attributes:
+                raise BundleError(_(
+                    "invalid when_creating attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
+                ).format(
+                    item=item_id,
+                    bundle=bundle.name,
+                    attrs=", ".join(invalid_attributes),
+                ))
 
     @classmethod
     def _validate_name(cls, bundle, name):
+        if not name:
+            raise BundleError(_(
+                "invalid name for {type} in bundle '{bundle}': must not be empty string"
+            ).format(
+                bundle=bundle.name,
+                type=cls.ITEM_TYPE_NAME,
+            ))
         if ":" in name:
             raise BundleError(_(
                 "invalid name for {type} in bundle '{bundle}': {name} (must not contain colon)"
@@ -423,8 +459,8 @@ class Item:
 
     def apply(
         self,
-        autoskip_selector=[],
-        autoonly_selector=[],
+        autoskip_selector=(),
+        autoonly_selector=(),
         my_soft_locks=(),
         other_peoples_soft_locks=(),
         interactive=False,
@@ -442,24 +478,6 @@ class Item:
         details = None
         start_time = datetime.now()
 
-        if not self.covered_by_autoonly_selector(autoonly_selector):
-            io.debug(_(
-                "autoonly does not match {item} on {node}"
-            ).format(item=self.id, node=self.node.name))
-            status_code = self.STATUS_SKIPPED
-            details = self.SKIP_REASON_CMDLINE
-
-        if self.covered_by_autoskip_selector(autoskip_selector):
-            io.debug(_(
-                "autoskip matches {item} on {node}"
-            ).format(item=self.id, node=self.node.name))
-            status_code = self.STATUS_SKIPPED
-            details = self.SKIP_REASON_CMDLINE
-
-        if self._skip_with_soft_locks(my_soft_locks, other_peoples_soft_locks):
-            status_code = self.STATUS_SKIPPED
-            details = self.SKIP_REASON_SOFTLOCK
-
         for item in self._precedes_items:
             if item._triggers_preceding_items(interactive=interactive):
                 io.debug(_(
@@ -472,21 +490,43 @@ class Item:
                     "preceding item {item} on {node} has NOT been triggered by {other_item}"
                 ).format(item=self.id, node=self.node.name, other_item=item.id))
 
-        if self.triggered and not self.has_been_triggered and status_code is None:
+        if self.skip:
+            status_code = self.STATUS_SKIPPED
+            details = self.SKIP_REASON_ATTR
+
+        elif self.triggered and not self.has_been_triggered:
             io.debug(_(
                 "skipping {item} on {node} because it wasn't triggered"
             ).format(item=self.id, node=self.node.name))
             status_code = self.STATUS_SKIPPED
             details = self.SKIP_REASON_NO_TRIGGER
 
-        if status_code is None and self.cached_unless_result and status_code is None:
+        elif not self.covered_by_autoonly_selector(autoonly_selector):
+            io.debug(_(
+                "autoonly does not match {item} on {node}"
+            ).format(item=self.id, node=self.node.name))
+            status_code = self.STATUS_SKIPPED
+            details = self.SKIP_REASON_CMDLINE
+
+        elif self.covered_by_autoskip_selector(autoskip_selector):
+            io.debug(_(
+                "autoskip matches {item} on {node}"
+            ).format(item=self.id, node=self.node.name))
+            status_code = self.STATUS_SKIPPED
+            details = self.SKIP_REASON_CMDLINE
+
+        elif self._skip_with_soft_locks(my_soft_locks, other_peoples_soft_locks):
+            status_code = self.STATUS_SKIPPED
+            details = self.SKIP_REASON_SOFTLOCK
+
+        elif self.cached_unless_result:
             io.debug(_(
                 "'unless' for {item} on {node} succeeded, not fixing"
             ).format(item=self.id, node=self.node.name))
             status_code = self.STATUS_SKIPPED
             details = self.SKIP_REASON_UNLESS
 
-        if self._faults_missing_for_attributes and status_code is None:
+        elif self._faults_missing_for_attributes:
             if self.error_on_missing_fault:
                 self._raise_for_faults()
             else:
@@ -503,7 +543,7 @@ class Item:
                 status_code = self.STATUS_SKIPPED
                 details = self.SKIP_REASON_FAULT_UNAVAILABLE
 
-        if status_code is None:
+        else:
             try:
                 status_before = self.cached_status
             except FaultUnavailable:
@@ -536,7 +576,7 @@ class Item:
                             sorted(copy(status_before.keys_to_fix)),
                         )
 
-        if status_code is None:
+        if status_code is None:  # item not skipped or OK
             if not interactive:
                 with io.job(_("{node}  {bundle}  {item}").format(
                     bundle=bold(self.bundle.name),
@@ -546,26 +586,17 @@ class Item:
                     self.fix(status_before)
             else:
                 if status_before.must_be_created:
-                    question_text = "\n".join(
-                        f"{bold(key)}  {value}"
-                        for key, value in sorted(details.items())
-                    )
+                    question_text = dict_to_text(details, value_color=green)
                     prompt = _("Create {}?").format(bold(self.id))
                 elif status_before.must_be_deleted:
-                    question_text = "\n".join(
-                        f"{bold(key)}  {value}"
-                        for key, value in sorted(details.items())
-                    )
+                    question_text = dict_to_text(details, value_color=red)
                     prompt = _("Delete {}?").format(bold(self.id))
                 else:
                     display_cdict, display_sdict, display_keys_to_fix = details
-                    question_text = "\n".join(
-                        diff_value(
-                            key,
-                            display_sdict[key],
-                            display_cdict[key],
-                        )
-                        for key in sorted(display_keys_to_fix)
+                    question_text = diff_dict(
+                        display_sdict,
+                        display_cdict,
+                        skip_missing_in_target=True,
                     )
                     prompt = _("Fix {}?").format(bold(self.id))
                 if self.comment:
@@ -598,7 +629,7 @@ class Item:
                     status_code = self.STATUS_SKIPPED
                     details = self.SKIP_REASON_INTERACTIVE
 
-        if status_code is None:
+        if status_code is None:  # item not skipped or OK
             status_after = self.get_status(cached=False)
             status_code = self.STATUS_FIXED if status_after.correct else self.STATUS_FAILED
 
@@ -647,7 +678,7 @@ class Item:
     def covered_by_autoskip_selector(self, autoskip_selector):
         """
         True if this item should be skipped based on the given selector
-        string (e.g. "tag:foo,bundle:bar").
+        (e.g. ("tag:foo", "bundle:bar")).
         """
         components = [c.strip() for c in autoskip_selector]
         if (
@@ -662,10 +693,10 @@ class Item:
                 return True
         return False
 
-    def covered_by_autoonly_selector(self, autoonly_selector):
+    def covered_by_autoonly_selector(self, autoonly_selector, check_deps=True):
         """
-        True if this item should be NOT skipped based on the given selector
-        string (e.g. "tag:foo,bundle:bar").
+        True if this item should NOT be skipped based on the given selector
+        (e.g. ("tag:foo", "bundle:bar")).
         """
         if not autoonly_selector:
             return True
@@ -679,16 +710,17 @@ class Item:
         for tag in self.tags:
             if "tag:{}".format(tag) in components:
                 return True
-        for depending_item in self._incoming_deps:
-            if (
-                depending_item.id in components or
-                "bundle:{}".format(depending_item.bundle.name) in components or
-                "{}:".format(depending_item.ITEM_TYPE_NAME) in components
-            ):
-                return True
-            for tag in depending_item.tags:
-                if "tag:{}".format(tag) in components:
+        if check_deps:
+            for depending_item in self._incoming_needs:
+                if (
+                    depending_item.id in components or
+                    "bundle:{}".format(depending_item.bundle.name) in components or
+                    "{}:".format(depending_item.ITEM_TYPE_NAME) in components
+                ):
                     return True
+                for tag in depending_item.tags:
+                    if "tag:{}".format(tag) in components:
+                        return True
         return False
 
     def fix(self, status):
@@ -745,7 +777,23 @@ class Item:
             return self.name
         return "{}:{}".format(self.ITEM_TYPE_NAME, self.name)
 
-    def verify(self):
+    def verify(
+        self,
+        autoskip_selector=(),
+        autoonly_selector=(),
+    ):
+        if not self.covered_by_autoonly_selector(autoonly_selector, check_deps=False):
+            io.debug(_(
+                "autoonly does not match {item} on {node}"
+            ).format(item=self.id, node=self.node.name))
+            raise ItemSkipped
+
+        if self.covered_by_autoskip_selector(autoskip_selector):
+            io.debug(_(
+                "autoskip matches {item} on {node}"
+            ).format(item=self.id, node=self.node.name))
+            raise ItemSkipped
+
         if self.cached_status.must_be_created:
             display = self.display_on_create(copy(self.cached_status.cdict))
         elif self.cached_status.must_be_deleted:
