@@ -16,7 +16,6 @@ from bundlewrap.exceptions import BundleError, FaultUnavailable, TemplateError
 from bundlewrap.items import BUILTIN_ITEM_ATTRIBUTES, Item
 from bundlewrap.items.directories import validator_mode
 from bundlewrap.utils import cached_property, hash_local_file, sha1, tempfile
-from bundlewrap.utils.dicts import diff_value_text
 from bundlewrap.utils.remote import PathInfo
 from bundlewrap.utils.text import force_text, mark_for_translation as _
 from bundlewrap.utils.text import is_subdirectory
@@ -64,7 +63,7 @@ def content_processor_jinja2(item):
             node=item.node.name,
         ))
     duration = datetime.now() - start
-    io.debug("{node}:{bundle}:{item}: rendered in {time}s".format(
+    io.debug("{node}:{bundle}:{item}: rendered in {time:.09f} s".format(
         bundle=item.bundle.name,
         item=item.id,
         node=item.node.name,
@@ -114,7 +113,7 @@ def content_processor_mako(item):
             node=item.node.name,
         ))
     duration = datetime.now() - start
-    io.debug("{node}:{bundle}:{item}: rendered in {time}s".format(
+    io.debug("{node}:{bundle}:{item}: rendered in {time:.09f} s".format(
         bundle=item.bundle.name,
         item=item.id,
         node=item.node.name,
@@ -254,12 +253,14 @@ class File(Item):
     def _fix_mode(self, status):
         if self.node.os in self.node.OS_FAMILY_BSD:
             command = "chmod {} {}"
+            mode = self.attributes['mode']
         else:
             command = "chmod {} -- {}"
-        self.run(command.format(
-            self.attributes['mode'],
-            quote(self.name),
-        ))
+            # GNU chmod refuses to set some modes (e.g., "chmod 0755 dir"
+            # when "dir" is a directory which currently has mode "2755")
+            # unless there's an additional leading zero.
+            mode = '0' + self.attributes['mode'].zfill(4)
+        self.run(command.format(mode, quote(self.name)))
 
     def _fix_owner(self, status):
         group = self.attributes['group'] or ""
@@ -347,7 +348,7 @@ class File(Item):
             len(self.content) < DIFF_MAX_FILE_SIZE
         ):
             del cdict['content_hash']
-            cdict['content'] = diff_value_text("", "", force_text(self.content)).rstrip("\n")
+            cdict['content'] = force_text(self.content)
         del cdict['type']
         return cdict
 
@@ -363,11 +364,19 @@ class File(Item):
             keys.append('content')
             del cdict['content_hash']
             del sdict['content_hash']
-            cdict['content'] = self.content
-            sdict['content'] = get_remote_file_contents(self.node, self.name)
+            cdict['content'] = self.content.decode(self.attributes['encoding'])
+            sdict['content'] = get_remote_file_contents(
+                self.node,
+                self.name,
+            ).decode('utf-8', 'backslashreplace')
         if 'type' in keys:
             with suppress(ValueError):
                 keys.remove('content_hash')
+        if sdict:
+            del sdict['size']
+            if self.attributes['content_type'] == 'any':
+                with suppress(KeyError):
+                    del sdict['content_hash']
         return (cdict, sdict, keys)
 
     def display_on_delete(self, sdict):
@@ -377,11 +386,7 @@ class File(Item):
             sdict['size'] < DIFF_MAX_FILE_SIZE and
             path_info.is_text_file
         ):
-            sdict['content'] = diff_value_text(
-                "",
-                get_remote_file_contents(self.node, self.name),
-                "",
-            ).rstrip("\n")
+            sdict['content'] = get_remote_file_contents(self.node, self.name)
         if path_info.is_file:
             sdict['size'] = f"{sdict['size']} bytes"
         return sdict
