@@ -1,12 +1,11 @@
 from contextlib import suppress
-from importlib.machinery import SourceFileLoader
+from importlib.util import module_from_spec, spec_from_file_location
 from inspect import isabstract
 from os import listdir, mkdir, walk
 from os.path import abspath, dirname, isdir, isfile, join
-from threading import RLock
 
 from pkg_resources import DistributionNotFound, require, VersionConflict
-from tomlkit import parse as parse_toml
+from tomlkit import parse as toml_parse
 
 from . import items, VERSION_STRING
 from .bundle import FILENAME_ITEMS
@@ -95,7 +94,7 @@ nodes = {
 }
 
 
-class HooksProxy(object):
+class HooksProxy:
     def __init__(self, repo, path):
         self.repo = repo
         self.__hook_cache = {}
@@ -166,19 +165,23 @@ class LibsProxy:
             filename = attrname + ".py"
             filepath = join(self.__path, filename)
             try:
-                m = SourceFileLoader(
+                spec = spec_from_file_location(
                     'bundlewrap.repo.libs_{}'.format(attrname),
                     filepath,
-                ).load_module()
+                )
+                mod = module_from_spec(spec)
+                spec.loader.exec_module(mod)
             except:
                 io.stderr(_("Exception while trying to load {}:").format(filepath))
                 raise
-            self.__module_cache[attrname] = m
+            self.__module_cache[attrname] = mod
         return self.__module_cache[attrname]
 
 
 class Repository(MetadataGenerator):
     def __init__(self, repo_path=None):
+        super().__init__()
+
         if repo_path is None:
             self.path = "/dev/null"
         else:
@@ -192,10 +195,6 @@ class Repository(MetadataGenerator):
         self._get_all_attr_code_cache = {}
         self._get_all_attr_result_cache = {}
 
-        # required by MetadataGenerator
-        self._node_metadata_proxies = {}
-        self._node_metadata_lock = RLock()
-
         if repo_path is not None:
             self.populate_from_path(self.path)
         else:
@@ -367,7 +366,7 @@ class Repository(MetadataGenerator):
                         not isfile(filepath) or \
                         filename.startswith("_"):
                     continue
-                infodict = dict(parse_toml(get_file_contents(filepath)))
+                infodict = toml_parse(get_file_contents(filepath)).value
                 infodict['file_path'] = filepath
                 yield filename[:-5], infodict
 
